# FILE: dialog.tk
#
# Tcl/Tk scripts for various dialog boxes in mmDisp
#

###################### OMF_PrintDialog ###################################
# 4-Feb-1997, mjd
#
# "Local" global variables are stored in array OMFpd$id(), where id is
# obtained by converting .'s to _'s in the $window value passed
# to OMF_PrintDialog

# Utility proc to derive a printing output filename from an
# input name.  Currently just strips any extension, and
# appends ".ps".
proc OMF_PrintDialogDefaultFilename { inname } {
    return "[file rootname [string trim $inname]].ps"
}

proc OMF_PrintDialogKill { window } {
    OMF_SetFocus . ;# Transfer focus back to root window
                   ## before destroying current window
    OMF_GetID $window id base
    destroy $base   ;# Important: You will be unable to delete
                     # $printvar until _after_ $base (or at least
                     # the entry widget tied to $printvar) is
                     # destroyed!
    set printvar "OMFpd$id"
    global $printvar
    trace vdelete $printvar w OMF_PrintDialogTrace
    global UpdateViewCount
    trace vdelete UpdateViewCount w OMF_PrintDialogViewportTrace$id
    unset $printvar
}

# Determines width:height ratio using values of 
# pvar(canvas) and pvar(crop)
proc OMF_PrintDialogSetSizeRatio { pvarname } {
    upvar #0 $pvarname pvar
    if {$pvar(crop)} {
        # Printing cropped (viewable) region
        set width [$pvar(canvas) cget -width]
        set height [$pvar(canvas) cget -height]
    } else {
        # Printing whole canvas
        set sr [.canvas_frame.canvas cget -scrollregion]
        set width [expr [lindex $sr 2] - [lindex $sr 0]]
        set height [expr [lindex $sr 3] - [lindex $sr 1]]
    }
    if {$width<=0}  { set width 1. }   ;# Safety: Both ratio &
    if {$height<=0} { set height 1. }  ;# 1/ratio should be valid.
    global $pvarname
    set ${pvarname}(sizeratio) [expr double($width)/double($height)]
    # We need to access pvar(sizeratio) as ${pvarname}(sizeratio)
    # in order to get the trace on ${pvarname} to work properly.
}

# Initialize OMF_PrintDialog data array from outside print array
proc OMF_PrintDialogCopyIn { window parrname } {
    OMF_GetID $window id
    set pvarname "OMFpd$id"
    upvar #0 $pvarname pvar
    upvar #0 $parrname parr
    set pvar(window)    $window
    set pvar(canvas)    $parr(canvas)
    set pvar(inname)    $parr(inname)
    set pvar(outname)   $parr(outname)
    set pvar(orient)    $parr(orient)
    set pvar(papertype) $parr(papertype)
    set pvar(crop)      $parr(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set pvar(verpos) {} ;# Default is centered
    regexp -nocase {[ns]} $parr(position) pvar(verpos)
    set pvar(horpos) {} ;# Default is centered
    regexp -nocase {[ew]} $parr(position) pvar(horpos)
    # Size is handled differently internal to OMF_PrintDialog:
    set pvar(units) "i"  ;# Safety
    regexp -nocase {[ic]} $parr(xmargin) pvar(units)
    set pvar(oldunits) $pvar(units) ;# Used to track unit conversions
    set pvar(leftmargin) [OMF_ConvertUnits $parr(xmargin) $pvar(units)]
    set pvar(topmargin) [OMF_ConvertUnits $parr(ymargin) $pvar(units)]
    OMF_PrintDialogSetSizeRatio $pvarname
    set direction width; set length 1i  ;# Safety
    if ![regexp -- {-page(width|height) +([0-9.]+.)} $parr(size) match \
            direction length] {
        Ow_FatalError "Invalid print size switch ($parr(size)) detected" \
                "OMF_PrintDialogCopyIn Error"
    }
    if {"$direction"=="width"} {
        set width  [OMF_ConvertUnits $length $pvar(units)]
        set height [expr $width/$pvar(sizeratio)]
    } else {
        set height [OMF_ConvertUnits $length $pvar(units)]
        set width  [expr $height*$pvar(sizeratio)]
    }

    # Strip extra precision off of width & height
    set width [string trim $width]
    if {[string compare $width "."]==0} {
        set width "0."
    }
    if [regexp {\....} $width] {
        set width [format "%.2f" $width]
    }

    set height [string trim $height]
    if {[string compare $height "."]==0} {
        set height "0."
    }
    if [regexp {\....} $height] {
        set height [format "%.2f" $height]
    }

    set pvar(width)  $width
    set pvar(height) $height

    # If outname doesn't specify a pipe, build a default name
    # from inname
    if { "|" != [string index [string trim $pvar(outname)] 0] } {
	# Not a pipe
	set pvar(outname) [OMF_PrintDialogDefaultFilename $pvar(inname)]
    }

}

# Copy OMF_PrintDialog data array to outside print array
proc OMF_PrintDialogCopyOut { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    OMF_GetID $window id
    set printvar "OMFpd$id"
    upvar #0 $printvar pvar
    upvar #0 $parrname parr
    set parr(inname)    $pvar(inname)
    set parr(outname)   $pvar(outname)
    set parr(orient)    $pvar(orient)
    set parr(papertype) $pvar(papertype)
    set parr(crop)      $pvar(crop)
    # Position is broken into two variables inside OMF_PrintDialog
    set parr(position)  "$pvar(verpos)$pvar(horpos)"
    if {"$parr(position)"==""} { set parr(position) "c" }

    # Size is handled differently internal to OMF_PrintDialog:
    set width [string trim $pvar(width)]
    if {[string compare $width ""]==0 \
        || [string compare $width "."]==0} {
        set width 0.
    }
    set parr(size)  "-pagewidth $width$pvar(units)"

    set leftmargin [string trim $pvar(leftmargin)]
    if {[string compare $leftmargin ""]==0 \
        || [string compare $leftmargin "."]==0} {
        set leftmargin 0.
    }
    set parr(xmargin) "$leftmargin$pvar(units)"

    set topmargin [string trim $pvar(topmargin)]
    if {[string compare $topmargin ""]==0 \
        || [string compare $topmargin "."]==0} {
        set topmargin "0."
    }
    set parr(ymargin) "$topmargin$pvar(units)"
}

# Fix up variables that depend on viewport display
proc OMF_PrintDialogUpdateVariables { pvarname } {
    OMF_PrintDialogSetSizeRatio $pvarname
}

# Trace procedure to automatically update PrintDialog variables
proc OMF_PrintDialogTrace { name elt op } {
    global $name
    if ![info exists ${name}($elt)] return
    # Disable tracing because we will be modifying several
    # elements of pvar.  Note: We must use the name "pvar"
    # inside the 'trace vdelete' command instead of "$name"
    # because that is how it is accessed inside this routine.
    upvar #0 $name pvar  ;# The #0 is needed for values
                          # modified inside entry widgets.
    trace vdelete pvar w OMF_PrintDialogTrace
    switch -exact -- $elt {
        sizeratio -
        width   {
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        height  {
            set temp [OMF_MakeValidPosNumber $pvar(height)]
            set pvar(height) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(width) ""
            } else {
                set pvar(width) [expr $temp*$pvar(sizeratio)]
            }
        }
        crop    {
            OMF_PrintDialogSetSizeRatio $name
            set temp [OMF_MakeValidPosNumber $pvar(width)]
            set pvar(width) $temp
            if {[string compare $temp ""]==0 \
                    || [string compare $temp "."]==0 } {
                set pvar(height) ""
            } else {
                set pvar(height) [expr $temp/$pvar(sizeratio)]
            }
        }
        units {
            set pvar(units) [string trim $pvar(units)] ;# Safety
            if {[string compare $pvar(units) $pvar(oldunits)]!=0} {
                set pvar(leftmargin) [OMF_ConvertUnits \
                        "$pvar(leftmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(topmargin) [OMF_ConvertUnits \
                        "$pvar(topmargin)$pvar(oldunits)" $pvar(units)]
                set pvar(width) [OMF_ConvertUnits \
                        "$pvar(width)$pvar(oldunits)" $pvar(units)]
                set pvar(height) [OMF_ConvertUnits \
                        "$pvar(height)$pvar(oldunits)" $pvar(units)]
                set pvar(oldunits) $pvar(units)
            }
        }
        default {}
    }
    # Strip extra precision off of width & height
    if [regexp {\....} $pvar(width)] {
        set pvar(width) [format "%.2f" $pvar(width)]
    }
    if [regexp {\....} $pvar(height)] {
        set pvar(height) [format "%.2f" $pvar(height)]
    }

    # Re-enable trace
    trace variable pvar w OMF_PrintDialogTrace
}


proc OMF_PrintDialogOK { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    OMF_GetID $window id
    set pvarname "OMFpd$id"
    upvar #0 $pvarname pvar
    upvar #0 $parrname parr

    # Don't allow exit if outname is {}.  Instead
    # reset outname to default name derived from inname,
    # and return.
    if { [string match {} $pvar(outname)] } {
	set pvar(outname) [OMF_PrintDialogDefaultFilename $pvar(inname)]
	$window.file.value xview end   ;# Make end of outname viewable
	return
    }

    # Otherwise, copy pvar to parr and kill Print Dialog
    OMF_PrintDialogCopyOut $window $parrname
    set parr(done) "OK"
    OMF_PrintDialogKill $window
    # Note: The outside trace on parr(done) is removed when the
    # PrintDialog window is destroyed, so it is necessary to write
    # "OK" to parr(done) *before* calling OMF_PrintDialogKill.
}

proc OMF_PrintDialogCancel { window parrname } {
    if { ![winfo exists $window] } { return }  ;# Window already destroyed
    upvar #0 $parrname parr
    set parr(done) "Cancel"
    OMF_PrintDialogKill $window
}

# Here $parrname must be a global variable
proc OMF_PrintDialog { window title parrname } {
    toplevel $window
    wm group $window .
    wm title $window "$title"
    Ow_PositionChild $window  ;# Position at +.25+.25 over '.'
    focus $window
    upvar #0 $parrname parr
    # parr is a *global* "Printer Array" structure:
    #    canvas     -   name of canvas widget
    #    inname     -   name of file being displayed
    #    outname    -   name of output file or pipe
    #    orient     -   portrait | landscape
    #    size       -   output width *or* height (not both), in the format
    #                     '-pagewidth 3.i'  (3 inches width), or
    #                     '-pageheight 5.c' (5 cm high), for example
    #    papertype  -   letter | legal | A4 | A3
    #    crop       -    1 | 0  (1=>print only displayed region of canvas)
    #    xmargin    -    left/right margin, with units.  E.g.: 1.i
    #    ymargin    -    top/bottom margin,  as above
    #    position   -    One of: c nw n ne e se s sw w

    # Packing frames
    set col1frame [frame $window.row1frame -bd 0]
    set col2frame [frame $window.row2frame -bd 0]
    set outputwidgets [frame $window.outputwidgets -relief ridge -bd 4]
    set sizewidgets   [frame $window.sizewidgets   -relief ridge -bd 4]
    set actionwidgets [frame $window.actionwidgets -relief ridge -bd 4]

    # Use $printvar() for "local" global variables
    OMF_GetID $window id
    set printvar OMFpd$id
    global $printvar
    OMF_PrintDialogCopyIn $window $parrname ;# Set initial values

    # Filename
    OMF_MakeEntryWindow $window.file "Print to:" {} \
            ${printvar}(outname) 15 0 raised
    $window.file.value xview end   ;# Make end of string viewable
    bind $window.file.value <Key-Escape> \
	    "set ${printvar}(outname) \{$parr(outname)\}"

    # Orientation
    OMF_MakeRadioList $window.orient 1 0 "Orientation:" {} \
            ${printvar}(orient) {} Portrait Landscape

    # Paper type
    OMF_MakeRadioList $window.papertype 2 0 "Paper type" {} \
            ${printvar}(papertype) {} Letter Legal A4 A3

    # Position
    OMF_MakeRadioList $window.horpos 2 0 "Horizontal Position" {} \
            ${printvar}(horpos) {} {Left w} {Center ""} {Right e}
    OMF_MakeRadioList $window.verpos 2 0 "Vertical Position" {} \
            ${printvar}(verpos) {} {Top n} {Center ""} {Bottom s}

    # Units
    OMF_MakeRadioList $window.units 1 1 "Units:" {} \
            ${printvar}(units) {} {in i} {cm c}

    # Margins
    frame $window.marginframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.topmargin "Margin Top:" 12 \
            ${printvar}(topmargin) 6 0 flat
    OMF_MakeEntryWindow $window.leftmargin "Left:" 7 \
            ${printvar}(leftmargin) 6 0 flat
    pack $window.topmargin $window.leftmargin \
            -side left -in $window.marginframe

    # Print dimensions
    frame $window.dimframe -relief ridge -bd 2
    OMF_MakeEntryWindow $window.width "Print Width:" 12 \
            ${printvar}(width) 6 0 flat
    OMF_MakeEntryWindow $window.height "Height:" 7 \
            ${printvar}(height) 6 0 flat
    pack $window.width $window.height -side left -in $window.dimframe

    # Crop?
    frame $window.cropframe -relief ridge -bd 2
    checkbutton $window.crop -text "Crop to view"  -relief flat \
            -variable ${printvar}(crop)
    pack $window.crop -side left -in $window.cropframe -expand 1

    # Action buttons
    button $window.ok -text "OK" -width 8 -command \
            "Oc_ThreadSafe OMF_PrintDialogOK,$window \
	    \{OMF_PrintDialogOK $window $parrname\}"
    button $window.reset -text "Reset" -width 8 -command \
            "Oc_ThreadSafe OMF_PrintDialogCopyIn,$window \
	    \{OMF_PrintDialogCopyIn $window $parrname\}"
    button $window.cancel -text "Cancel" -width 8 -command \
            "Oc_ThreadSafe OMF_PrintDialogCancel,$window \
	    \{OMF_PrintDialogCancel $window $parrname\}"
    wm protocol $window WM_DELETE_WINDOW "$window.cancel invoke"

    # Pack
    pack $window.file $window.orient $window.papertype \
            $window.horpos $window.verpos \
            -side top -fill x -expand 1 -in $outputwidgets 
    pack $window.units $window.marginframe $window.dimframe \
            $window.cropframe \
            -side top -fill x -expand 1 -in $sizewidgets
    pack $window.ok $window.reset $window.cancel -side left -expand 1 \
            -in $actionwidgets

    pack $outputwidgets -side top -fill x -in $col1frame \
            -padx 5 -pady 5 -expand 1 -anchor n
    pack $sizewidgets   -side top -fill x -in $col2frame \
            -padx 5 -pady 5 -anchor n
    pack $actionwidgets -side top -in $col2frame -fill x \
             -padx 5 -pady 5 -expand 1 -anchor s
    pack $col1frame -side left -fill both -expand 1
    pack $col2frame -side left -fill y

    # Trace to keep printvar internally consistent
    trace variable $printvar w OMF_PrintDialogTrace

    # Trace to update printvar if viewport in . changes
    # This is a little tricky, because we need to send the
    # name $printvar through the trace (which is not tied
    # to $printvar, but to global variable UpdateViewCount.
    # There is probably a cleaner way to do this...  -mjd 97/2/11
    proc OMF_PrintDialogViewportTrace$id { name elt op } \
            "OMF_PrintDialogUpdateVariables $printvar"
    global UpdateViewCount
    trace variable UpdateViewCount w OMF_PrintDialogViewportTrace$id

    Ow_SetIcon $window
    focus $window.file.value
}



###################### Plot Select and Configuration #####################
# 4-June-1998, mjd
#
# Local configuration results are stored in array psc.
Oc_Class Mmd_PlotSelectConfigure {
    const public variable title = "mmDisp: Configure"
    const public variable winpath

    # Import trace and initialization array.  Must be global.
    const public variable import_arrname = {}

    # Routine in mainline to call on an Apply or OK event.
    public variable apply_callback = Oc_Nop 

    # Menu and index (as a 2-item list) in the launching window
    # to toggle off in the constructor and on in the destructor.
    const public variable menu = {}

    # Internal state array
    private array variable psc

    # State (normal/disabled) dependency lists
    private array variable state_dep

    # Listbox update lists array
    private array variable listbox_dep

    Constructor {args} {
        # Create a (presumably) unique name for the toplevel
        # window based on the instance name.
        set winpath ".[$this WinBasename]"

	# Process user preferences
	eval $this Configure $args

        # Create dialog window
        toplevel $winpath
        wm group $winpath .
        wm title $winpath $title
        Ow_PositionChild $winpath ;# Position at +.25+/25 over '.'
        focus $winpath

        # Setup destroy binding.
	bind $winpath <Destroy> "$this WinDestroy %W"

        # Disable launch menu, if given
        if {![string match {} $menu]} {
            [lindex $menu 0] entryconfigure [lindex $menu 1] \
                    -state disabled
        }

        # Setup internal state array default values
        array set psc {
            colormaps  { Red-Black-Blue Blue-White-Red \
                    Black-Gray-White White-Green-Black }
	    quantitylist       {}
            arrow,status       1
            arrow,colormap     Red-Black-Blue
            arrow,colorcount   256
            arrow,quantity     z
            arrow,subsample    {}
            arrow,mag          1
            pixel,status       0
            pixel,colormap     Red-Black-Blue
            pixel,colorcount   256
            pixel,quantity     x
            pixel,subsample    0
            pixel,mag          1
            misc,drawboundary  1
            misc,background    white
            misc,margin        10
        }
        set gpsc [$this GlobalName psc]
        ## Global name of psc array.  This is needed to tie
        ## elements of this array to Tk buttons.

	set listbox_dep(colormaps) {}
	set listbox_dep(quantitylist) {}

        # Allow override from user initialization array
        if {![string match {} $import_arrname]} {
            upvar #0 $import_arrname userinit
            if {[info exists userinit]} {
                foreach elt [array names userinit] {
                    if {[info exists psc($elt)]} {
                        set psc($elt) $userinit($elt)
                    }
                }
            }
            trace variable userinit w "$this ExternalUpdate $import_arrname"
        }

        ########### Create dialog box
        # Grid Cells
        for {set i 0} {$i<3} {incr i} {
            for {set j 0} {$j<6} {incr j} {
                set f [frame $winpath.a$i$j -bd 2 -relief ridge]
                set a$i$j $f
                grid $f -row $i -column $j -sticky news
            }
        }

        # Column labels
        set clbl0 [label $a00.label -text "Plot Type"]
        set clbl1 [label $a01.label -text "Colormap"]
        set clbl2 [label $a02.label -text "# of Colors"]
        set clbl3 [label $a03.label -text "Color Quantity"]
        set clbl4 [label $a04.label -text "Subsample Rate"]
        set clbl5 [label $a05.label -text "Magnification"]

        pack $clbl0 $clbl1 $clbl2 $clbl3 $clbl4 $clbl5 -fill both

        # Arrow plot configuration
	set state_dep(arrow) {}
        set arrstat [checkbutton $a10.stat -text "Arrow" \
                -variable ${gpsc}(arrow,status) \
                -onvalue 1 -offvalue 0]
        $arrstat configure -command "$this UpdatePlotTypeState arrow"
        pack $arrstat -fill both

        set disabledcolor [$arrstat cget -disabledforeground] ;# Used
        ## below in Ow_EntryBox configuration

        Ow_ListBox New arrcm $a11.cm -height 3 \
                -variable ${gpsc}(arrow,colormap) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $arrcm Insert end $psc(colormaps)
        pack $a11.cm -anchor e
        lappend state_dep(arrow) $arrcm
	lappend listbox_dep(colormaps) $arrcm

	Ow_EntryBox New arrcc $a12.cc -fixpt 1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,colorcount)
        grid [$arrcc Cget -winpath]
        lappend state_dep(arrow) $arrcc

        Ow_ListBox New arrqt $a13.qt -height 3 \
                -variable ${gpsc}(arrow,quantity) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $arrqt Insert end $psc(quantitylist)
        pack $a13.qt -fill both -expand 1
        lappend state_dep(arrow) $arrqt
	lappend listbox_dep(quantitylist) $arrqt

        Ow_EntryBox New arrsr $a14.sr -fixpt -1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,subsample)
        grid [$arrsr Cget -winpath]
        lappend state_dep(arrow) $arrsr

        Ow_EntryBox New arrmag $a15.mag -fixpt -1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(arrow,mag)
        grid [$arrmag Cget -winpath]
        lappend state_dep(arrow) $arrmag

        # Initialize arrow type display state
        $this UpdatePlotTypeState arrow

        # Pixel plot configuration
	set state_dep(pixel) {}
        set pixstat [checkbutton $a20.stat -text "Pixel" \
                -variable ${gpsc}(pixel,status) -onvalue 1 -offvalue 0]
        $pixstat configure -command "$this UpdatePlotTypeState pixel"
        pack $pixstat -fill both

        Ow_ListBox New pixcm $a21.cm -height 3 \
                -variable ${gpsc}(pixel,colormap) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $pixcm Insert end $psc(colormaps)
        pack $a21.cm -anchor w
        lappend state_dep(pixel) $pixcm
	lappend listbox_dep(colormaps) $pixcm

	Ow_EntryBox New pixcc $a22.cc -fixpt 1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,colorcount)
        grid [$pixcc Cget -winpath]
        lappend state_dep(pixel) $pixcc

        Ow_ListBox New pixqt $a23.qt -height 3 \
                -variable ${gpsc}(pixel,quantity) \
                -outer_frame_options "-bd 0" \
		-listbox_options "-bd 2 -relief flat -width 0"
        eval $pixqt Insert end $psc(quantitylist)
        pack $a23.qt -fill both -expand 1
        lappend state_dep(pixel) $pixqt
	lappend listbox_dep(quantitylist) $pixqt

        Ow_EntryBox New pixsr $a24.sr -fixpt -1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,subsample)
        grid [$pixsr Cget -winpath]
        lappend state_dep(pixel) $pixsr

        Ow_EntryBox New pixmag $a25.mag -fixpt -1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label {} -valuewidth 5 \
                -disabledforeground $disabledcolor \
                -variable ${gpsc}(pixel,mag)
        grid [$pixmag Cget -winpath]
        lappend state_dep(pixel) $pixmag
    
        # Initialize pixel type display state
        $this UpdatePlotTypeState pixel

        # Miscellaneous configuration
        set a30 [frame $winpath.a30 -bd 2 -relief ridge]
        grid $a30 -row 3 -column 0 -columnspan 6 -sticky news
        set f30 [frame $a30.f30 -bd 0]
        pack $f30 -side left -fill y
        
        set bdrycb [checkbutton $f30.bdrycb -text "Draw Boundary" \
                -variable ${gpsc}(misc,drawboundary) -onvalue 1 -offvalue 0]

        Ow_EntryBox New marginbox $f30.margin -fixpt -1 \
                -outer_frame_options "-bd 0 -relief flat" \
                -label "Margin:" -valuewidth 5 \
                -variable ${gpsc}(misc,margin)                

        grid $bdrycb x [$marginbox Cget -winpath] - -sticky w

        grid rowconfigure $winpath 3 -pad 10
        set bgclab [label $f30.bgclab -text "Background color:"]
        set bgc0 [radiobutton $f30.bgc0 -text "white" \
                -variable ${gpsc}(misc,background) -value "white"]
        set bgc1 [radiobutton $f30.bgc1 -text "black" \
                -variable ${gpsc}(misc,background) -value "black"]
        set bgc2 [radiobutton $f30.bgc2 -text "green" \
                -variable ${gpsc}(misc,background) -value "green"]
        grid $bgclab $bgc0 $bgc1 $bgc2 -sticky w

        # Control buttons
        set a40 [frame $winpath.a40 -bd 0]
        grid $a40 -row 4 -column 0 -columnspan 6 -sticky news
        grid rowconfigure $winpath 4 -pad 10
        set cancel [button $a40.cancel -text "Cancel" \
                -command "$this Action cancel" ]
        set apply [button $a40.apply -text "Apply" \
                -command "$this Action apply" ]
        set ok [button $a40.ok -text "OK" \
                -command "$this Action ok" ]
        pack $cancel $apply $ok -side left -expand 1
        wm protocol $winpath WM_DELETE_WINDOW "$cancel invoke"
    }

    # Update from external source
    callback method ExternalUpdate { globalarrname name elt ops } {
        ## The trace is setup to pass in the global array name so that
        ## even if the trace is triggered by a call to Tcl_SetVar
        ## in C code called from a proc from which the traced array
        ## is not visible (due to no 'global' statement on that array),
        ## this trace can still function.        
        if {[string match {} $elt]} {return}
        upvar #0 $globalarrname source
        if {[info exists psc($elt)] && \
                ![string match $psc($elt) $source($elt)]} {
            set psc($elt) $source($elt)
	    # Reset listboxes as necessary
	    if {[info exists listbox_dep($elt)]} {
		foreach lb $listbox_dep($elt) {
		    $lb Remove 0 end
		    eval {$lb Insert end} $psc($elt)		    
		}
	    }
        }
    }

    # Plot type status change helper function
    callback method UpdatePlotTypeState { plottype } {
        if {$psc($plottype,status)} {
            set mystate normal
        } else {
            set mystate disabled
        }
        foreach win $state_dep($plottype) {
            $win configure -state $mystate
        }
    }

    # Dialog action routine
    callback method Action { cmd } {
        switch $cmd {
            cancel {
                $this Delete
                return
            }
            apply {
                eval $apply_callback psc
            }
            ok {
                eval $apply_callback psc
                $this Delete
                return
            }
            default {
                error "Illegal action code: $cmd"
            }
        }
    }

    # Destructor methods
    method WinDestroy { w } { winpath } {
	if {[string compare $winpath $w]!=0} { return } ;# Child destroy event
	$this Delete
    }
    Destructor {
        # Re-enable launching menu
        if {![string match {} $menu]} {
            if {[catch { [lindex $menu 0] entryconfigure \
                    [lindex $menu 1] -state normal } errmsg]} {
            Oc_Log Log $errmsg error
            }
        }
        # Remove external update trace, if any
        if {[catch {
            if {![string match {} $import_arrname]} {
                upvar #0 $import_arrname userinit
                trace vdelete userinit w "$this ExternalUpdate $import_arrname"
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
        # Destroy windows
	if {[catch { 
            if {[info exists winpath] && [winfo exists $winpath]} {
                # Remove <Destroy> binding
                bind $winpath <Destroy> {}
                # Destroy toplevel, all children, and all bindings
                destroy $winpath
            }
        } errmsg]} {
            Oc_Log Log $errmsg error
        }
    }
}
