/* FILE: display.cc                 -*-Mode: c++-*-
 *
 * C++ code for display windows
 *
 */

#include <stdlib.h>
#include <string.h>

#include "display.h"
#include "constants.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Colormap class
const ClassDoc DisplayColorMap::class_doc("DisplayColorMap",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","15-Dec-1996");

#if (TKCOLORLENGTH != 6)
# error "Current code only supports TKCOLORLENGTH==6"
#endif

void DisplayColorMap::Free()
{
  if(TkColor==NULL) return;
  delete[] TkColor[0];
  delete[] TkColor;
  TkColor=NULL;
}

void DisplayColorMap::Alloc(int _colorcount)
{
  if(TkColor!=NULL) Free();
  if(_colorcount<1) colorcount=1;
  else              colorcount=_colorcount;
  TkColor=new char*[colorcount];
  TkColor[0]=new char[colorcount*(TKCOLORLENGTH+2)];
  /// Leave room for '#' & '\0'
  for(int i=1;i<colorcount;i++) {
    TkColor[i]=TkColor[i-1]+TKCOLORLENGTH+2;
  }
}

void DisplayColorMap::Setup(int _colorcount,const char* _mapname,
			    const char* MakeColor(REAL4m relative_shade))
{
  mapname=_mapname;
  if(MakeColor==NULL) {
    Alloc(1);
    strncpy(TkColor[0],"#000000",TKCOLORLENGTH+2);
    return;
  }
  Alloc(_colorcount);
  if(colorcount==1) {
    strncpy(TkColor[0],MakeColor(0.),TKCOLORLENGTH+2);
  } else {
    for(int i=0;i<colorcount;i++) {
      REAL4m relshade=double(i)/double(colorcount-1);
      strncpy(TkColor[i],MakeColor(relshade),TKCOLORLENGTH+2);
      /// Include room for '#' & '\0'
    }
  }
}

DisplayColorMap::DisplayColorMap(void)
{
#define MEMBERNAME "DisplayColorMap"
  TkColor=NULL;
  Setup(0,"unknown",NULL);
#undef MEMBERNAME
}

DisplayColorMap::DisplayColorMap(const DisplayColorMap& colormap)
{
  TkColor=NULL;
  *this=colormap;
}

DisplayColorMap& DisplayColorMap::operator=(const DisplayColorMap& colormap)
{
  Alloc(colormap.colorcount);
  for(int i=0;i<colorcount;i++) {
    strncpy(TkColor[i],colormap.GetColor(i),TKCOLORLENGTH+2);
    /// Include room for '#' & '\0'
  }
  return *this;
}

const char* DisplayColorMap::GetColor(REAL4m shade) const
{ // Import shade should in [0.,1.]
  int index=(int)OC_ROUND(shade*REAL4m(colorcount-1));
  if(index<0)                index=0;
  else if(index>=colorcount) index=colorcount-1;
  return TkColor[index];
}

const char* DisplayColorMap::GetColor(INT4m index) const
{ // Retrieves color based on index.
  if(index<0) index=0;
  if(index>=colorcount) index=colorcount-1;
  return TkColor[index];
}

void DisplayColorMap::Setup(int _colorcount,const char* _mapname)
{ // A setup routine using the predefined MakeColor routines below.
#define MEMBERNAME "Setup(int,const char*)"
  if(strcmp(_mapname,"Red-Black-Blue")==0) {
    Setup(_colorcount,_mapname,MakeColorRedBlackBlue);
  } else if(strcmp(_mapname,"Blue-White-Red")==0) {
    Setup(_colorcount,_mapname,MakeColorBlueWhiteRed);
  } else if(strcmp(_mapname,"Black-Gray-White")==0) {
    Setup(_colorcount,_mapname,MakeColorBlackGrayWhite);
  } else if(strcmp(_mapname,"White-Green-Black")==0) {
    Setup(_colorcount,_mapname,MakeColorWhiteGreenBlack);
  } else {
    FatalError(-1,STDDOC,"Unknown mapname: %s",_mapname);
  }
#undef MEMBERNAME
}

// A few sample MakeColor functions.  These are mapped through
// an arccos, which will display a component map as linear in
// angle.
const char* MakeColorRedBlackBlue(REAL4m relative_shade)
{
  static char color[TKCOLORLENGTH+3]; // 1 for safety
  double x=acos(1-2*relative_shade);
  int value=(int)floor(256*x/PI);
  if(value<0)   value=0;
  if(value>255) value=255;
  int red,green,blue;
  red=blue=0; // To keep dumb compilers quiet
  green=0;
  if(value<128) {
    blue=0;
    red=255-2*value;
  } else if(value>=128) {
    red=0;
    blue=2*(value-128)+1;
  }
  sprintf(color,"#%02X%02X%02X",red,green,blue);
  color[TKCOLORLENGTH+2]='\0'; // Safety
  return color;
}

const char* MakeColorBlueWhiteRed(REAL4m relative_shade)
{
  static char color[TKCOLORLENGTH+3]; // 1 for safety
  double x=acos(1-2*relative_shade);
  int value=(int)floor(256*x/PI);
  if(value<0)   value=0;
  if(value>255) value=255;
  int red,green,blue;
  red=green=blue=0; // To keep dumb compilers quiet
  if(value<128) {
    blue=255;
    red=green=2*value;
  } else if(value>=128) {
    red=255;
    blue=green=510-2*value;
  }
  sprintf(color,"#%02X%02X%02X",red,green,blue);
  color[TKCOLORLENGTH+2]='\0'; // Safety
  return color;
}

const char* MakeColorBlackGrayWhite(REAL4m relative_shade)
{
  static char color[TKCOLORLENGTH+3]; // 1 for safety
  double x=acos(1-2*relative_shade);
  int value=(int)floor(256*x/PI);
  if(value<0)   value=0;
  if(value>255) value=255;
  sprintf(color,"#%02X%02X%02X",value,value,value);
  color[TKCOLORLENGTH+2]='\0'; // Safety
  return color;
}

const char* MakeColorWhiteGreenBlack(REAL4m relative_shade)
{
  static char color[TKCOLORLENGTH+3]; // 1 for safety
  double x=acos(1-2*relative_shade);
  int value=(int)floor(256*x/PI);
  if(value<0)   value=0;
  if(value>255) value=255;
  int red,green,blue;
  red=blue=255-value;
  green=0; // To keep dumb compilers quiet
  if(value<128) green=255;
  else          green=510-2*value;
  sprintf(color,"#%02X%02X%02X",red,green,blue);
  color[TKCOLORLENGTH+2]='\0'; // Safety
  return color;
}


//////////////////////////////////////////////////////////////////////////
// PlotConfiguration structure
//
const ClassDoc PlotConfiguration::class_doc("PlotConfiguration",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.1","16-Jul-1998");
PlotConfiguration::PlotConfiguration()
{
  displaystate=1;
  subsample=0.0;
  mag=1.0;
  colormap.Setup(256,"Red-Black-Blue");

}

PlotConfiguration& PlotConfiguration::operator=(const PlotConfiguration& pc)
{
  displaystate=pc.displaystate;
  subsample=pc.subsample;
  mag=pc.mag;
  colorquantity=pc.colorquantity;
  colormap=pc.colormap;
  return *this;
}

PlotConfiguration::PlotConfiguration(const PlotConfiguration& pc)
{
  *this=pc;
}

//////////////////////////////////////////////////////////////////////////
// CoordinateSystem helper function
INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle)
{ // Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
  // both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
  // returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
  // is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
  // In these latter two cases angle is 0.0 on return.
  double ang1,ang2;
  angle=0.0;
  switch(cs1) {
  case CS_DISPLAY_STANDARD: ang1=  0; break;
  case CS_DISPLAY_ROT_90:   ang1= 90; break;
  case CS_DISPLAY_ROT_180:  ang1=180; break;
  case CS_DISPLAY_ROT_270:  ang1=270; break;
  default: return 1;  // Illegal coordinates
    // break;
  }
  switch(cs2) {
  case CS_DISPLAY_STANDARD: ang2=  0; break;
  case CS_DISPLAY_ROT_90:   ang2= 90; break;
  case CS_DISPLAY_ROT_180:  ang2=180; break;
  case CS_DISPLAY_ROT_270:  ang2=270; break;
  default: return 2;  // Illegal coordinates
    // break;
  }
  angle=ang1-ang2;
  if(angle<0) angle+=360;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Display frame class
const ClassDoc DisplayFrame::class_doc("DisplayFrame",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","14-Dec-1996");
// Recommended display window width, height and cellsize, in pixels:
const REAL4m DisplayFrame::default_arrow_cellsize=20;
const REAL4m DisplayFrame::default_pixel_cellsize=4;
const REAL4m DisplayFrame::default_arrow_ratio=0.85;
const REAL4m DisplayFrame::default_pixel_ratio=1.00;

void DisplayFrame::SetPreferredArrowCellsize(REAL4m pixels)
{
  if(pixels>0. && preferred_arrow_cellsize!=pixels) {
    preferred_arrow_cellsize=pixels;
    if(arrow_config.subsample<0) arrow_valid=0;
  }
}

void DisplayFrame::SetPreferredPixelCellsize(REAL4m pixels)
{
  if(pixels>0. && preferred_pixel_cellsize!=pixels) {
    preferred_pixel_cellsize=pixels;
    if(pixel_config.subsample<0) pixel_valid=0;
  }
}

DisplayFrame::DisplayFrame()
{
  mesh=(Mesh *)NULL;
  tcl_interp=Oc_GlobalInterpreter();
  coords=CS_DISPLAY_STANDARD;
  arrow_valid=pixel_valid=boundary_valid=0;
  mesh_per_pixel=1.0;
  background_color="white";
  preferred_arrow_cellsize=default_arrow_cellsize;
  preferred_pixel_cellsize=default_pixel_cellsize;
}

void DisplayFrame::ChangeCoordinates(CoordinateSystem old_coords,
				     CoordinateSystem new_coords,
				     List<DisplayVector> &display_list) const
{
#define MEMBERNAME "ChangeCoordinates(CS,CS,List<DisplayVector> &"
  if(old_coords==new_coords) return;
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
  REAL4m pixel_per_mesh=1.0/mesh_per_pixel;
  Vec3f mesh_min,mesh_max;
  BoundingBox mesh_range;
  mesh->GetRange(mesh_range);
  mesh_range.GetExtremes(mesh_min,mesh_max);

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(DisplayVector* display_vector=display_list.GetFirst();
	display_vector!=(DisplayVector *)NULL;
	display_vector=display_list.GetNext()) {
      // The "->location" section of this could be implemented
      // by a call to ConvertMeshPtToDisplayPt, but this was written
      // before that routine was, and this should be faster for
      // a long list.
      display_vector->location.x -= mesh_min.x;
      display_vector->location.y  = mesh_max.y - display_vector->location.y;
      display_vector->location.z -= mesh_min.z;
      display_vector->location*=pixel_per_mesh;
      display_vector->value.y*=-1;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeCoordinates(old_coords,CS_DISPLAY_STANDARD,display_list);
    ChangeCoordinates(CS_DISPLAY_STANDARD,new_coords,display_list);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(OC_ROUND(dangle_diff));
    Vec3f display_max=mesh_max,vtemp;
    display_max-=mesh_min;
    display_max*=pixel_per_mesh;
    for(DisplayVector* display_vector=display_list.GetFirst();
	display_vector!=(DisplayVector *)NULL;
	display_vector=display_list.GetNext()) {
      Vec3f *loc=&(display_vector->location);
      Vec3f *dir=&(display_vector->value);
      switch(angle_diff) {
      case 90:
	vtemp = *loc;
	loc->x = vtemp.y;
	loc->y = display_max.y - vtemp.x;
	vtemp = *dir;
	dir->x =  vtemp.y;
	dir->y = -vtemp.x;
	break;
      case 180:
	loc->x = display_max.x - loc->x;
	loc->y = display_max.y - loc->y;
	dir->x *= -1;
	dir->y *= -1;
	break;
      case 270:
	vtemp = *loc;
	loc->x = display_max.x - vtemp.y;
	loc->y = vtemp.x;
	vtemp = *dir;
	dir->x = -vtemp.y;
	dir->y =  vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeCoordinates(CoordinateSystem old_coords,
				     CoordinateSystem new_coords,
				     List<Vec3f> &pt_list) const
{ // Same as previous, but for locations only
#define MEMBERNAME "ChangeCoordinates(CS,CS,List<Vec3f> &"
  if(old_coords==new_coords) return;
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
  REAL4m pixel_per_mesh=1.0/mesh_per_pixel;
  Vec3f mesh_min,mesh_max;
  BoundingBox mesh_range;
  mesh->GetRange(mesh_range);
  mesh_range.GetExtremes(mesh_min,mesh_max);

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(Vec3f* pt=pt_list.GetFirst();pt!=(Vec3f *)NULL;
	pt=pt_list.GetNext()) {
      pt->x -= mesh_min.x;
      pt->y  = mesh_max.y - pt->y;
      pt->z -= mesh_min.z;
      (*pt) *= pixel_per_mesh;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeCoordinates(old_coords,CS_DISPLAY_STANDARD,pt_list);
    ChangeCoordinates(CS_DISPLAY_STANDARD,new_coords,pt_list);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff,temp;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(OC_ROUND(dangle_diff));
    Vec3f display_max=mesh_max;
    display_max-=mesh_min;
    display_max*=pixel_per_mesh;
    for(Vec3f* pt=pt_list.GetFirst() ; pt!=(Vec3f *)NULL;
	pt=pt_list.GetNext()) {
      switch(angle_diff) {
      case 90:
	temp=pt->x;
	pt->x=pt->y;
	pt->y=display_max.y-temp;
	break;
      case 180:
	pt->x = display_max.x - pt->x;
	pt->y = display_max.y - pt->y;
	break;
      case 270:
	temp=pt->x;
	pt->x = display_max.x - pt->y;
	pt->y = temp;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

INT4m DisplayFrame::GetDisplayList(SampleInfo &si)
{ // Fills display_list from *mesh using requested step size and
  // color quantity. Use step_request=0 to get _all_ points in mesh.
  // Converts to current display_coords.
  // NOTE: Samples mesh in terms of *mesh cells*, as opposed to
  //       *mesh units*.
  // NOTE: mesh_per_pixel must be set before calling this routine.
#define MEMBERNAME "GetDisplayList"
  if(si.subsample<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,si.subsample);
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Get mesh bounding box
  BoundingBox mesh_box;
  mesh->GetRange(mesh_box);

  // Determine stepping rates.  The code below is awkward; It is a
  // result of legacy code/design (or non-design, as the case may be).
  REAL4m mesh_cellsize=mesh->GetApproximateCellWidth();
  if(mesh_cellsize<=0.0) mesh_cellsize=1.0; // Safety
  Vec3f mesh_celldim=mesh->GetApproximateCellDimensions();
  if(mesh_celldim.x<=0.0) mesh_celldim.x=1.0; //Safety
  if(mesh_celldim.y<=0.0) mesh_celldim.y=1.0; //Safety

  // Get list of vectors to display
  REAL4m xstep=mesh_celldim.x*si.subsample;
  REAL4m ystep=mesh_celldim.y*si.subsample;
  REAL4m zstep=0.;
  mesh->GetDisplayList(xstep,ystep,zstep,mesh_box,
		       si.colorquantity,si.dvlist);
  xstep/=mesh_celldim.x;
  ystep/=mesh_celldim.y;
  si.subsample=OC_MAX(xstep,ystep);
  REAL4m grit=mesh->GetSubsampleGrit()*mesh_cellsize;
  if(0.<si.subsample && si.subsample<grit/2.) si.subsample=grit;

  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,si.dvlist);

  return 0;
#undef MEMBERNAME

}


void DisplayFrame::SetPlotConfiguration(
                const PlotConfiguration& _arrow_config,
                const PlotConfiguration& _pixel_config)
{
  // Copy arrow configuration update arrow display list
  arrow_config=_arrow_config;
  arrow_valid=0;

  // Ditto for pixel configuration
  pixel_config=_pixel_config;
  pixel_valid=0;
}

REAL4m DisplayFrame::GetAutoSampleRate(REAL4m cellsize) const
{ // Returns a recommended subsample rate, for the current mesh at
  // current zoom.
  REAL4m step=cellsize*mesh_per_pixel;
  step=mesh->GetPreferredStepSize(step);
  step/=mesh->GetApproximateCellWidth();
  step=OC_ROUND(step*100.)/100.;  // Make "round" value
  if(step<=1.) step=0.;
  return step;
}

INT4m DisplayFrame::DrawPixel(const char *canvas_name,
			      const DisplayVector &dspvec,
                              REAL4m pixelsize) const
{ // Based on code contributed by SLW
#define MEMBERNAME "DrawPixel"
  static char buf[1024];  // *Should* be plenty big enough.
  const Vec3f *postemp=&(dspvec.location);
  const Vec3f *vtemp=&(dspvec.value);
  if(vtemp->MagSq()<EPSILON) return TCL_OK;  // This is a zero vector

  // Color
  const char *color=pixel_config.Color(dspvec.shade);

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing %f
  //       conversions.  It is 6 or 7 times faster to use %d's instead.

  REAL4m x1,x2,y1,y2;
  int r1,r2,c1,c2;

  x1=(postemp->x)-pixelsize/2;  x2=x1+pixelsize;
  r1=int(floor(x1));        r2=int(ceil(x2));
  y1=(postemp->y)-pixelsize/2;  y2=y1+pixelsize;
  c1=int(floor(y1));        c2=int(ceil(y2));
  sprintf(buf,"%s create rect %d %d %d %d -fill %s"
          " -outline {} -tag pixels",
          canvas_name,r1,c1,r2,c2,color);

  return Tcl_Eval(tcl_interp,buf);
#undef MEMBERNAME
}

REAL8m DisplayFrame::GetPixelSize() const {
  REAL8m pixelsize=mesh->GetApproximateCellWidth()
    * default_pixel_ratio * pixel_config.mag / mesh_per_pixel;
  if(pixel_sample.subsample>1.)
    pixelsize*=pixel_sample.subsample;
  return pixelsize;
}

REAL8m DisplayFrame::GetArrowSize() const {
  REAL8m arrowsize=mesh->GetApproximateCellWidth()
    * default_arrow_ratio * arrow_config.mag / mesh_per_pixel;
  if(arrow_sample.subsample>1.)
    arrowsize*=arrow_sample.subsample;
  return arrowsize;
}

INT4m DisplayFrame::DrawArrow(const char *canvas_name,
			      const DisplayVector &dspvec,
                              REAL4m arrowsize) const
{
#define MEMBERNAME "DrawArrow"
  static char buf[4096];  // *Should* be plenty big enough.
  const Vec3f *postemp;
  const Vec3f *vtemp;

  REAL4m x1,y1,x2,y2,delx,dely,vec_mag,arrow_len,proj_len;
  INT4m r0,c0,r1,c1,r2,c2,ioff,a1,a2,a3;
  postemp=&(dspvec.location);
  vtemp=&(dspvec.value);
  // Line components
  delx=vtemp->x*arrowsize;    dely=vtemp->y*arrowsize;
  vec_mag=vtemp->Mag();
  arrow_len=vec_mag*arrowsize;
  if(arrow_len<SQRT_EPSILON) return TCL_OK;  // This is a zero vector
  proj_len=sqrt(delx*delx+dely*dely);
  // Arrowhead components
  a1=(int)OC_ROUND(proj_len*0.3);
  a2=(int)OC_ROUND(proj_len*0.4);
  a3=(int)OC_ROUND(arrow_len*0.15);
  if(a3<5) a1=a2;
  // Attributes
  int linewidth=(int)OC_ROUND(a3*0.6);
  if(linewidth<1) linewidth=1;

  const char* pcolor1=arrow_config.Color(dspvec.shade);

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing
  //       %f conversions.  It is 6 or 7 times faster to use %d's instead
  //      of sprintf(buf,"%s create line %6.2f %6.2f %6.2f %6.2f"
  //	             "-arrow last -fill %s",canvas_name,x1,y1,x2,y2,
  //                 colormap.TkColor[color]);

  if(proj_len>=arrow_len*0.15) { // Standard arrow
    x1=(postemp->x)-delx/2;   x2=x1+delx;
    r1=int(OC_ROUND(x1));     r2=int(OC_ROUND(x2));
    y1=(postemp->y)-dely/2;   y2=y1+dely;
    c1=int(OC_ROUND(y1));     c2=int(OC_ROUND(y2));
    sprintf(buf,
            "%s create line %d %d %d %d -tag arrows "
            "-arrow last -arrowshape {%d %d %d}"
            "  -width %d -fill %s",canvas_name,r1,c1,r2,c2,a1,a2,a3,
            linewidth,pcolor1);
    /// Note: The return value from sprintf is not consistent across
    ///  all systems.  ANSI says "int," but SunOS returns "char *".
  }
  else { // Small projection with undefined direction.  Draw a dot
    REAL4m backshade=dspvec.shade+0.5;  // Maybe there's a better way
    backshade-=floor(backshade);       /// to do this?
    const char* pcolor2=arrow_config.Color(backshade);
    r0=(int)OC_ROUND(postemp->x); c0=(int)OC_ROUND(postemp->y);
    if(vtemp->z<0) { // Arrow pointing into-plane
      ioff=(int)OC_ROUND(REAL4m(a3)/2.);
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(buf,"%s create rect %d %d %d %d -fill %s -outline %s",
	      canvas_name,r1,c1,r2,c2,pcolor1,pcolor1);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(buf,'\0');
	ioff=(int)OC_ROUND(REAL4m(a3)/6.);
	r1=r0-ioff;  c1=c0-ioff;
	r2=r0+ioff;  c2=c0+ioff;
	sprintf(bufptr,"; %s create rect %d %d %d %d -fill %s -width 0",
		canvas_name,r1,c1,r2,c2,pcolor2);
      }
    } else { // Arrow pointing out-of-plane
      ioff=(int)OC_ROUND(REAL4m(a3)*(SQRT2/2.));
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(buf,
	      "%s create poly %d %d %d %d %d %d %d %d"
	      " -fill %s -outline %s -width %d",
	      canvas_name,r1,c0,r0,c2,r2,c0,r0,c1,
	      pcolor1,pcolor1,linewidth/3);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(buf,'\0');
	sprintf(bufptr,"; %s create line %d %d %d %d -fill %s -width %d ; "
		"%s create line %d %d %d %d -fill %s -width %d ; ",
		canvas_name,r1,c0,r2,c0,pcolor2,linewidth/3,
		canvas_name,r0,c2,r0,c1,pcolor2,linewidth/3);
      }
    }
  }
  if(Tcl_Eval(tcl_interp,buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawLine(const char *canvas_name,
			     const Vec3f &Pt1,const Vec3f &Pt2)
{
  static char buf[1024];  // *Should* be plenty big enough.
  sprintf(buf,
	  "%s create line %6.2f %6.2f %6.2f %6.2f",
	  canvas_name,Pt1.x,Pt1.y,Pt2.x,Pt2.y);
  if(Tcl_Eval(tcl_interp,buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
}

const char* DisplayFrame::GetVecColor(const Vec3f& v) const
{
  // Access to the underlying Mesh::GetVecShade function.  See that
  // routine for details on shade determination.  The mapping from
  // shade to color depends on the state of arrow_config.displaystate
  // and pixel_config.displaystate.  Pixel config colorquantity and
  // colormap are used, unless pixels are turned off and arrows are
  // turned on, in which case arrrow config colorquantity and colormap
  // are used.
  REAL4m shade;
  const char* color="";
  if(!pixel_config.displaystate && arrow_config.displaystate) {
    // Use arrow_config data
    shade=mesh->GetVecShade(arrow_config.colorquantity,v);
    if(shade>=0.) color=arrow_config.Color(shade);
  } else {
    // Use pixel_config data
    shade=mesh->GetVecShade(pixel_config.colorquantity,v);
    if(shade>=0.) color=pixel_config.Color(shade);
  }
  return color;
}

INT4m DisplayFrame::Show(const char *canvas_name)
{
#define MEMBERNAME "Show"
  // Initialize canvas
  if(tcl_interp==NULL)
    FatalError(-1,STDDOC,"tcl_interp not initialized.");
  if(mesh==NULL)
    FatalError(-1,STDDOC,"mesh not initialized.");

  // Check that current coordinate system is displayable.
  if(coords!=CS_DISPLAY_STANDARD &&
     coords!=CS_DISPLAY_ROT_90   &&
     coords!=CS_DISPLAY_ROT_180  &&
     coords!=CS_DISPLAY_ROT_270    ) {
    FatalError(-1,STDDOC,
	       "Can't draw display list in current (%d) coordinate system",
	       coords);
  }

  // Set canvas background color
  const char *fmt="%s configure -background %s";
  char *buf=new char[strlen(fmt)+strlen(canvas_name)
		    +background_color.Length()+1];
  sprintf(buf,"%s configure -background %s",canvas_name,
	  background_color.GetStr());
  if(Tcl_Eval(tcl_interp,buf) != TCL_OK) return TCL_ERROR;
  delete[] buf;

  // Draw pixel map, if requested
  if(pixel_config.displaystate) {
    if(!pixel_valid) {
      pixel_sample.subsample=pixel_config.subsample;
      if(pixel_sample.subsample<0.)
	pixel_sample.subsample=GetAutoSampleRate(preferred_pixel_cellsize);
      pixel_sample.colorquantity=pixel_config.colorquantity;
      GetDisplayList(pixel_sample);
      pixel_valid=1;
    }
    REAL8m pixelsize=GetPixelSize();
    DisplayVector *dspvec;
    for(dspvec=pixel_sample.dvlist.GetFirst();
	dspvec!=(DisplayVector *)NULL;
	dspvec=pixel_sample.dvlist.GetNext()) {
      if(DrawPixel(canvas_name,*dspvec,pixelsize) != TCL_OK)
	return TCL_ERROR;
    }
  }

  // Draw arrows, if requested
  if(arrow_config.displaystate) {
    if(!arrow_valid) {
      arrow_sample.subsample=arrow_config.subsample;
      if(arrow_sample.subsample<0.)
	arrow_sample.subsample=GetAutoSampleRate(preferred_arrow_cellsize);
      arrow_sample.colorquantity=arrow_config.colorquantity;
      GetDisplayList(arrow_sample);
      arrow_valid=1;
    }
    REAL8m arrowsize=GetArrowSize();
    DisplayVector *dspvec;
    for(dspvec=arrow_sample.dvlist.GetFirst();
	dspvec!=(DisplayVector *)NULL;
	dspvec=arrow_sample.dvlist.GetNext()) {
      if(DrawArrow(canvas_name,*dspvec,arrowsize) != TCL_OK)
	return TCL_ERROR;
    }
  }

  // Draw mesh edge box
  if(draw_boundary) {
    if(!boundary_valid) {
      mesh->GetBoundaryList(boundary_list);
      ChangeCoordinates(CS_CALCULATION_STANDARD,coords,boundary_list);
      boundary_valid=1;
    }
    Vec3f *bdry1,*bdry2;
    if((bdry1=boundary_list.GetFirst())!=(Vec3f *)NULL) {
      while((bdry2=boundary_list.GetNext())!=(Vec3f *)NULL) {
	if(DrawLine(canvas_name,*bdry1,*bdry2) != TCL_OK)
	  return TCL_ERROR;
	bdry1=bdry2;
      }
    }
  }

  return TCL_OK;
#undef MEMBERNAME
}

// NOTE: "zoom" in the next several routines is in units of "pixels per
//       approx cell size".
REAL4m DisplayFrame::GetZoom(void) const
{
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  return mesh_per_cell/mesh_per_pixel;
}

REAL4m DisplayFrame::SetZoom(REAL4m zoom)
{ // Returns zoom value
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety

  if(EPSILON<zoom && zoom<1/EPSILON) {
    mesh_per_pixel=mesh_per_cell/zoom;
    arrow_valid=pixel_valid=boundary_valid=0;
  }

  return mesh_per_cell/mesh_per_pixel;
}

REAL4m DisplayFrame::SetZoom(REAL4m width,REAL4m height)
{ // Adjusts mesh_per_pixel scaling factor so that entire
  // mesh will just fit inside width and height.  Marks display
  // lists as out-of-date.  New zoom value is returned.
  // NOTE: Setting either width or height <=0 will cause
  //  that extent to be ignored.  If both are <=0, then
  //  no change is made.

  REAL8m new_mesh_per_pixel=mesh_per_pixel;

  if(new_mesh_per_pixel<=0.) new_mesh_per_pixel=1.0; // Just to be safe.

  BoundingBox bdry,data;
  REAL8m data_pad;

  GetBoundingBoxes(bdry,data,data_pad);

  REAL8m xscale= -1.0, yscale= -1.0;
  if(width>0.) {
    xscale=bdry.GetWidth()/width;
    if(width>2*data_pad)
      xscale=OC_MAX(xscale,data.GetWidth()/(width-2*data_pad));
  }
  if(height>0.) {
    yscale=bdry.GetHeight()/height;
    if(height>2*data_pad)
      yscale=OC_MAX(yscale,data.GetHeight()/(height-2*data_pad));
  }
  REAL8m scale=OC_MAX(xscale,yscale);

  if(scale>0.) new_mesh_per_pixel*=scale;

  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  REAL4m zoom=mesh_per_cell/new_mesh_per_pixel; // User value

  // Make zoom a "round" (%.2f) value
  if(zoom>0.01) {
    zoom=OC_ROUND(zoom*100.)/100.;
    /// Arguably this should be "floor" instead of "round," but the
    /// latter appears more stable with respect to dependence of
    /// outcome on initial value of mesh_per_pixel.  Moreover, there
    /// is now (June-1998) some slack in the calculation of whether
    /// scroll bars are required.
    new_mesh_per_pixel=double(mesh_per_cell)/double(zoom);
  }

  if(mesh_per_pixel!=REAL4m(new_mesh_per_pixel)) {
    mesh_per_pixel=new_mesh_per_pixel;
    MarkAllSampleListsInvalid();
  }

  return zoom;
}

void DisplayFrame::GetBoundingBoxes(BoundingBox &bdry,
				    BoundingBox &data,REAL8m &data_pad) const
{ // Returns bounding boxes for display, in pixels (at current zoom):
  //   'bdry' is the bounding box for the boundary
  //   'data' is the bounding box for the data
  //   'data_pad' is the margin (in pixels) that needs to be added
  //     to the data bounding box.
  // The first two scale with mesh_per_pixel (i.e., current zoom).  The
  // last is independent of the zoom levels.
  // NOTE: The returned BoundingBox'es are guaranteed to have max>=min.

  Vec3f min,max;

  // Check for an empty mesh
  BoundingBox bbox;
  mesh->GetRange(bbox);
  if(bbox.IsEmpty()) {
    // Note: Empty boxes may have max<min.  Change this
    // for return so that max==min.
    bbox.GetExtremes(min,max);
    max=min;
    data.Set(min,max);
    bdry.Set(min,max);
    data_pad=0.;
    return;
  }

  // Calculate bounding box for data points
  mesh->GetDataRange(data);
  data.GetExtremes(min,max);
  List<Vec3f> datalist;
  datalist.SetFirst(min); datalist.Append(max);
  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,datalist);
  min= *(datalist.GetFirst()); max= *(datalist.GetNext());
  data.SortAndSet(min,max);

  data_pad=OC_MAX(preferred_arrow_cellsize,preferred_pixel_cellsize)/2.;

  // Create boundary bounding box by expanding for each point in the
  // boundary list, one at a time.  Use supplemental list (instead of
  // boundary_list) so we don't have to worry about updating boundary_list
  // and loosing the 'const' property of this method.
  List<Vec3f> bdrylist;
  mesh->GetBoundaryList(bdrylist);
  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,bdrylist);
  bdry.Reset();
  for(Vec3f* bp=bdrylist.GetFirst() ; bp!=NULL ; bp=bdrylist.GetNext()) {
    bdry.ExpandWith(*bp);
  }
}

BoundingBox DisplayFrame::GetDisplayBox() const
{ // Returns bounding box for display, in pixels (at current zoom)
  BoundingBox bdry,data,total_box;
  REAL8m data_pad;

  GetBoundingBoxes(bdry,data,data_pad);

  // Add padding to data box.
  Vec3f min,max;
  data.GetExtremes(min,max);
  min.x-=data_pad; min.y-=data_pad; min.z-=data_pad;
  max.x+=data_pad; max.y+=data_pad; max.z+=data_pad;
  total_box.Set(min,max);

  // Expand as necessary to include boundary.
  total_box.ExpandWith(bdry);

  return total_box;
}








