/* FILE: vector.cc                -*-Mode: c++-*-
 *
 * Vector and collections of vectors
 *
 */

#include <stdio.h>
#include <string.h>

#include "vector.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// 4-byte floating point vectors
const ClassDoc Vec3f::class_doc("Vec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

//////////////////////////////////////////////////////////////////////////
// 3D rectangular solid
const ClassDoc BoundingBox::class_doc("BoundingBox",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","13-Dec-1996");

BOOL BoundingBox::IsIn(const Vec3f &location) const
{
#define MEMBERNAME "IsIn"
  return (location.Bigger(bbmin) && (bbmax.Bigger(location)));
#undef MEMBERNAME
}

void BoundingBox::SortAndSet(const Vec3f &corner1,const Vec3f &corner2) {
  if(corner1.x<corner2.x) { bbmin.x=corner1.x; bbmax.x=corner2.x; }
  else                    { bbmin.x=corner2.x; bbmax.x=corner1.x; }
  if(corner1.y<corner2.y) { bbmin.y=corner1.y; bbmax.y=corner2.y; }
  else                    { bbmin.y=corner2.y; bbmax.y=corner1.y; }
  if(corner1.z<corner2.z) { bbmin.z=corner1.z; bbmax.z=corner2.z; }
  else                    { bbmin.z=corner2.z; bbmax.z=corner1.z; }
}

void BoundingBox::Set(const Vec3f &corner1,const Vec3f &corner2) {
  bbmin.x=corner1.x; bbmax.x=corner2.x;
  bbmin.y=corner1.y; bbmax.y=corner2.y;
  bbmin.z=corner1.z; bbmax.z=corner2.z;
}

BOOL BoundingBox::ExpandWith(const Vec3f &include_point)
{ // Expands box as necessary to include 'include_point'. Returns
  // 1 if expansion occurred, 0 otherwise.
  if(IsEmpty()) {
    bbmin=bbmax=include_point;
    return 1;
  }
  BOOL expand=0;
  if(include_point.x<bbmin.x) { bbmin.x=include_point.x; expand=1; }
  if(include_point.y<bbmin.y) { bbmin.y=include_point.y; expand=1; }
  if(include_point.z<bbmin.z) { bbmin.z=include_point.z; expand=1; }
  if(include_point.x>bbmax.x) { bbmax.x=include_point.x; expand=1; }
  if(include_point.y>bbmax.y) { bbmax.y=include_point.y; expand=1; }
  if(include_point.z>bbmax.z) { bbmax.z=include_point.z; expand=1; }
  return expand;
}

BOOL BoundingBox::ExpandWith(const BoundingBox& other_box)
{ // Expands box as necessary to include 'other_box'.  Returns 1 if
  // expansion occurred, 0 otherwise
  if(IsEmpty()) {
    if(other_box.IsEmpty()) return 0;
    Set(other_box);
    return 1;
  }
  BOOL expand=0;
  if(other_box.bbmin.x<bbmin.x) { bbmin.x=other_box.bbmin.x; expand=1; }
  if(other_box.bbmin.y<bbmin.y) { bbmin.y=other_box.bbmin.y; expand=1; }
  if(other_box.bbmin.z<bbmin.z) { bbmin.z=other_box.bbmin.z; expand=1; }
  if(other_box.bbmax.x>bbmax.x) { bbmax.x=other_box.bbmax.x; expand=1; }
  if(other_box.bbmax.y>bbmax.y) { bbmax.y=other_box.bbmax.y; expand=1; }
  if(other_box.bbmax.z>bbmax.z) { bbmax.z=other_box.bbmax.z; expand=1; }
  return expand;
}

BoundingBox IntersectBoundingBoxes(const BoundingBox& box1,
				   const BoundingBox& box2)
{ // Returns the box resulting from the intersection of box1 and box2.
  // If the two imports have empty intersection, then the returned
  // box has .bbmin=(0,0,0), .bbmax=(-1,-1,-1)
  BoundingBox box;
  box.Reset();  // Safety
  if(box1.IsEmpty() || box2.IsEmpty()) return box;
  Vec3f bbmin,bbmax;
  bbmin.x=OC_MAX(box1.bbmin.x,box2.bbmin.x);
  bbmin.y=OC_MAX(box1.bbmin.y,box2.bbmin.y);
  bbmin.z=OC_MAX(box1.bbmin.z,box2.bbmin.z);
  bbmax.x=OC_MIN(box1.bbmax.x,box2.bbmax.x);
  bbmax.y=OC_MIN(box1.bbmax.y,box2.bbmax.y);
  bbmax.z=OC_MIN(box1.bbmax.z,box2.bbmax.z);
  box.Set(bbmin,bbmax);
  if(box.IsEmpty()) box.Reset();
  return box;
}

Vec3f BoundingBox::GetDiag(void) const { 
  if(IsEmpty()) return Vec3f(0.,0.,0.);
  return bbmax-bbmin;
}
