/* FILE: vector.h                 -*-Mode: c++-*-
 *
 * Vector and collections of vectors
 *
 */

#ifndef VECTOR_H
#define VECTOR_H

#include "template.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// 4-byte floating point vectors
class Vec3f
{
private:
  static const ClassDoc class_doc;
public:
  REAL4 x,y,z;  // Components
  void Set(REAL4 nx,REAL4 ny,REAL4 nz) { x=nx; y=ny; z=nz; }
  Vec3f(void) { x=y=z=0.0; } // Default constructor
  Vec3f(REAL4 nx,REAL4 ny,REAL4 nz) { Set(nx,ny,nz); }
  Vec3f(const Vec3f &v) : x(v.x), y(v.y), z(v.z) {} // Copy constructor
  // Assignment operator
  Vec3f& operator=(const Vec3f &v) { x=v.x; y=v.y; z=v.z; return *this; }
  BOOL Equals(const Vec3f &v) const { return (x==v.x  && y==v.y && z==v.z); }
  BOOL Bigger(const Vec3f &v) const { return (x>=v.x  && y>=v.y && z>=v.z); }
  inline REAL4 MagSq() const { return x*x+y*y+z*z; }
  inline REAL4 Mag() const { return sqrt(MagSq()); }
  inline Vec3f & operator+=(const Vec3f& v);
  inline Vec3f & operator-=(const Vec3f& v);
  inline Vec3f & operator*=(REAL4 a);
  friend Vec3f operator+(const Vec3f& v,const Vec3f& w);
  friend Vec3f operator-(const Vec3f& v,const Vec3f& w);
  friend Vec3f operator*(REAL4 a,const Vec3f& v);
  friend Vec3f operator*(const Vec3f& v,REAL4 a);
};

Vec3f & Vec3f::operator+=(const Vec3f& v) {
  x+=v.x; y+=v.y; z+=v.z; return *this;
}
Vec3f & Vec3f::operator-=(const Vec3f& v) {
  x-=v.x; y-=v.y; z-=v.z; return *this;
}
Vec3f & Vec3f::operator*=(REAL4 a)  { x*=a; y*=a; z*=a; return *this; }

inline Vec3f operator+(const Vec3f& v,const Vec3f& w)
{
  Vec3f temp;
  temp.x=v.x+w.x;  temp.y=v.y+w.y;  temp.z=v.z+w.z;
  return temp;
}

inline Vec3f operator-(const Vec3f& v,const Vec3f& w)
{
  Vec3f temp;
  temp.x=v.x-w.x;  temp.y=v.y-w.y;  temp.z=v.z-w.z;
  return temp;
}

inline Vec3f operator*(REAL4 a,const Vec3f& v)
{
  Vec3f temp;
  temp.x=a*v.x;  temp.y=a*v.y;  temp.z=a*v.z;
  return temp;
}

inline Vec3f operator*(const Vec3f& v,REAL4 a) { return a*v; }


//////////////////////////////////////////////////////////////////////////
// 4-byte floating point Position _and_ Vector
struct LocatedVector
{
public:
  Vec3f location;
  Vec3f value;
  LocatedVector() { location.Set(0.,0.,0.); value.Set(0.,0.,0.); }
  LocatedVector(const Vec3f &newloc,const Vec3f &newval) {
    location=newloc; value=newval;
  }
};

//////////////////////////////////////////////////////////////////////////
// Bounding Box
class BoundingBox
{ // Note: Empty boxes are well-defined.  See Empty() function below.
  static const ClassDoc class_doc;
  Vec3f bbmin;
  Vec3f bbmax;
public:
  BOOL IsEmpty() const { 
    return (bbmin.x>bbmax.x || bbmin.y>bbmax.y || bbmin.z>bbmax.z);
  }
  void Reset() { bbmin.Set(0.,0.,0.); bbmax.Set(-1.,-1.,-1.); }
  void SortAndSet(const Vec3f &corner1,const Vec3f &corner2);
  void Set(const Vec3f &corner1,const Vec3f &corner2);
  void Set(const BoundingBox &_other) { Set(_other.bbmin,_other.bbmax); }
  BOOL ExpandWith(const Vec3f &include_point); // Expands box as necessary to
  /// include 'include_point'.  Returns 1 if expansion occurred, 0 otherwise
  BOOL ExpandWith(const BoundingBox &other_box); // Expands box as necessary
  /// to include 'other_box'.  Returns 1 if expansion occurred, 0 otherwise
  friend BoundingBox IntersectBoundingBoxes(const BoundingBox& box1,
					    const BoundingBox& box2);
  /// Returns the box resulting from the intersection of box1 and box2.
  /// If the two imports have empty intersection, then the returned
  /// box has .bbmin=(0,0,0),.bbmax=(-1,-1,-1)
  
  // Default constructor
  BoundingBox(void) { Reset(); }
  // Copy constructor
  BoundingBox(const BoundingBox& newbb) :
    bbmin(newbb.bbmin), bbmax(newbb.bbmax) {}

  // Construction from two Vec3f's.
  BoundingBox(const Vec3f& v1,const Vec3f v2) : bbmin(v1), bbmax(v2) {}

  // Assignment operator
  BoundingBox& operator=(const BoundingBox& bb) { 
    bbmin=bb.bbmin; bbmax=bb.bbmax; return *this;
  }
  BOOL IsIn(const Vec3f &location) const;
  void GetMinPt(Vec3f &minpt) const { minpt=bbmin; }
  void GetMaxPt(Vec3f &maxpt) const { maxpt=bbmax; }
  void GetExtremes(Vec3f &minpt,Vec3f &maxpt) const {
    minpt=bbmin; maxpt=bbmax;
  }
  REAL4m GetWidth(void)  const { return OC_MAX(bbmax.x-bbmin.x,0.); }
  REAL4m GetHeight(void) const { return OC_MAX(bbmax.y-bbmin.y,0.); }
  REAL4m GetDepth(void)  const { return OC_MAX(bbmax.z-bbmin.z,0.); }
  REAL4m GetVolume(void) const { return GetWidth()*GetHeight()*GetDepth(); }
  Vec3f GetDiag(void) const; // Returns max-min if box is not empty,
  /// 0 otherwise.
};

#endif  // VECTOR_H
