#!/bin/sh
# FILE: batchslave.tcl
#   This script brings up a Tcl interpreter that hooks up to a network
# socket and then processes Tcl commands sent over that socket.  The
# return value of each command is sent back across the same socket when
# the command is complete.
#
# Note: If batchsolve.tcl is to be used with this script, then this
#  script should be launched with mmsolve instead of tclsh.
#
# Bootstrap: this comment continues on in Tcl, but not in sh \
exec tclsh "$0" ${1+"$@"}

if {[llength $argv]<4} {
    puts stderr "Usage: $argv0 server_host\
	    server_port slaveid passwd \[aux_script \[aux_args\]\]"
    puts stderr "(Got: $argv0 $argv)"
    exit
}

set base_dir [info script]
set root_oommf [file join [file dirname $base_dir] .. .. ..]
set script_path [list . ./scripts ./data \
	$base_dir \
	[file join $base_dir scripts] \
	[file join $base_dir data] \
	[file join $root_oommf app mmsolve] \
	[file join $root_oommf app mmsolve scripts] \
	[file join $root_oommf app mmsolve data]]

proc bgerror { msg } {
   global errorInfo
   puts stderr "BGERROR: $msg"
   puts stderr "STACK: $errorInfo"
}

# Global variables
global masterhost masterport ;# Host machine and port of master
global sock                  ;# Socket to master

# Message procs.  These escape newlines.
proc MyPuts { chan msg } {
    regsub -all -- "%" $msg "%25" msg      ;# Escape %'s
    regsub -all -- "\n" $msg "%10" msg     ;# Escape newlines
    puts $chan $msg
}
proc MyGets { chan msgname } {
    upvar $msgname msg
    set count [gets $chan msg]
    if {$count<0} {
	return $count    ;# Error or line not ready
    }
    regsub -all -- "%10" $msg "\n" msg     ;# Un-escape newlines
    regsub -all -- "%25" $msg "%" msg      ;# Un-escape %'s
    return [string length $msg]
}

# Exit handler
proc Die {} {
    global sock
    puts stderr "Slave dying on \
	    $sock=[lrange [fconfigure $sock -sockname] 1 end]"
    catch {MyPuts $sock "status bye"}
    catch {close $sock}
    exit
}

# This routine gets called to eval a command coming from
# the master.  The socket should be in non-blocking mode.
# The command is a 5 item list, of the form
#
#              request eval $taskno $idstring $script
#
# After task completion, the taskno, idstring, errorcode and result
# are sent back across the socket, via a
#
#              reply eval $taskno $idstring $errorcode $result
#
# message, followed by
#
#              status ready
#
# to indicate readiness to accept a new command.
proc ProcessCommand {} {
    global sock
    if {[catch {MyGets $sock line} count] || $count<0} {
	if {[eof $sock]} { Die } ;# Connection closed
	return  ;# Otherwise, input line not yet complete
    }
    if {[regexp -- "^\[ \t\]*\$" $line]} { return } ;# Ignore blank lines
    if {[catch {lindex $line 0} msggrp]} {
	puts stderr "Bad input line on channel $sock: $line"
	return
    }
    if {![string match "request" $msggrp]} {
	puts stderr "Unexpected input on channel $sock: $line"
	return
    }
    set msgcmd [lindex $line 1]
    if {[string match "die" $msgcmd] || [string match "bye" $msgcmd]} {
	Die
    }
    if {![string match "eval" $msgcmd]} {
	puts stderr "Unexpected input on channel $sock: $line"
	return
    }
    set taskno [lindex $line 2]
    set idstring [lindex $line 3]
    set cmd [lindex $line 4]
    set errorcode [catch {uplevel #0 $cmd} result]  ;# Eval at global scope
    MyPuts $sock [list reply eval $taskno $idstring $errorcode $result]
    MyPuts $sock "status ready"
}

# Hook up to master
set masterhost [lindex $argv 0]
set masterport [lindex $argv 1]
if {[catch {socket $masterhost $masterport} sock]} {
    puts stderr "Error opening socket: $sock"
    exit
}
fconfigure $sock -buffering line -blocking 0
fileevent $sock readable ProcessCommand
MyPuts $sock "connect verify [lindex $argv 2] [lindex $argv 3]"

# Source in auxiliary script, if any
if {[llength $argv] > 4} {
    set orig_argv $argv
    set argv [lrange $argv 5 end] ;# Fake argument list
    set aux_script [lindex $orig_argv 4]
    foreach dir $script_path {
	set attempt [file join $dir $aux_script]
	if {[file readable $attempt]} {
	    set aux_script $attempt
	    break
	}
    }
    if {[catch {source $aux_script} msg]} {
	puts stderr "Error sourcing [lindex $orig_argv 4]: $msg"
    }
    set argv $orig_argv
}

# Inform master we are ready to begin processing, and go into event loop
after 20 {catch {MyPuts $sock "status ready"}}

vwait forever
