# FILE: log.tcl
#
#	General logging facility.
#
# Last modified on: $Date: 1998/07/15 04:11:39 $
# Last modified by: $Author: dgp $
#
# This file defines the Tcl command Oc_Log for directing messages from
# multiple sources and in multiple categories to multiple handlers for
# recording or display.
#
# This facility was inspired by a similar facility in Pool, a library
# of Tcl routines by Andreas Kupries <a.kupries@westend.com> which is
# available at <URL:http://www.westend.com/%7Ekupries/doc/pool/index.htm>.
#
# DOME: Consider procs for deleting Types and Sources.

Oc_Class Oc_Log {
    array common db {, ""}

    proc AddType {type} {db} {
        if {[regexp , $type]} {
            return -code error \
                    "The character ',' is not allowed in a message type."
        }
        if {![info exists db(,$type)]} {
            foreach _ [array names db *,] {
                regexp ^(.*),$ $_ _ src
                array set db [list $src,$type $db($src,)]
            }
        }
    }

    proc AddSource {src} {db} {
        if {[regexp , $src]} {
            return -code error \
                    "The character ',' is not allowed in a message source."
        }
        if {![info exists db($src,)]} {
            foreach _ [array names db ,*] {
                regexp ^,(.*)$ $_ _ type 
                array set db [list $src,$type $db(,$type)]
            }
        }
    }

    proc SetLogHandler {hdlr type {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        set db($src,$type) $hdlr
    }

    proc GetLogHandler {type {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        return $db($src,$type)
    }

    proc Log {msg {type ""} {src ""}} {db} {
        if {![info exists db($src,$type)]} {
            return -code error "Unknown (source,type) pair: ($src,$type)"
        }
        if {[string match {} $db($src,$type)]} {
            # No handler, do nothing
            return
        }

        # Arguments quoted by [list] so that uplevel doesn't break them apart.
        if {[catch {
                uplevel #0 $db($src,$type) [list $msg] [list $type] [list $src]
                } ret]} {
            if {[string match Oc_log $src] && [string match error $type]} {
                # Prevent infinite loops
                return
            }
            global errorInfo errorCode
            error "Error returned by LogHandler '$db($src,$type)':
    	$ret

	while handling message of type '$type' from source '$src':
	$msg" $errorInfo $errorCode
        }
    }
}

Oc_Log AddType panic	;# Fatal errors
Oc_Log AddType error	;# Non-fatal errors reported via Tcl's error unwinding
Oc_Log AddType warning ;# Non-fatal errors reported some other way
			;# (like from deep within the C++ code)
Oc_Log AddType info 	;# A noteworthy, unusual event occurred, 
			;# but not an error.
Oc_Log AddType status	;# The passing of a routine event
			;# (like an iteration count)
Oc_Log AddType debug	;# Detailed messages useful only for development.

# Redefine bgerror to use Oc_log
proc bgerror {msg} {
    global errorCode errorInfo

    if {[string match OC [lindex $errorCode 0]]} {
        Oc_Log Log $msg error [lindex $errorCode 1]
    } else {
        Oc_Log Log $msg error
    }
}

# Redefine tclLog (Tcl 8.0+) to use Oc_log
proc tclLog {string} {
    Oc_Log Log $string warning
}

# Set up default destinations for most important log messages
if {[Oc_Main HasTk]} {
    proc Oc_DefaultMessage {msg type src} {
        tk_dialog .message "[pid] $src $type:" $msg error 0 OK
    }
} else {
    if {[string match windows $tcl_platform(platform)] \
            && [llength [info commands Oc_WindowsMessageBox]]} {
        proc Oc_DefaultMessage {msg type src} {
            Oc_WindowsMessageBox "[pid] $src $type: $msg"
        }
    } else {
        proc Oc_DefaultMessage {msg type src} {
            puts stderr "[pid] $src $type: $msg"
            flush stderr
        }
    }
}

Oc_Log SetLogHandler Oc_DefaultMessage panic Oc_Log
Oc_Log SetLogHandler Oc_DefaultMessage error Oc_Log
Oc_Log SetLogHandler Oc_DefaultMessage warning Oc_Log
Oc_Log SetLogHandler Oc_DefaultMessage panic
Oc_Log SetLogHandler Oc_DefaultMessage error
Oc_Log SetLogHandler Oc_DefaultMessage warning
Oc_Log AddSource Oc_Class

